import { LimitedCache } from 'limited-cache';
import React, { useMemo, useRef, useState } from 'react';
import { isElement } from 'react-is';
import { InPortal, OutPortal, createHtmlPortalNode } from 'react-reverse-portal';
import { Route, Redirect, Switch } from 'react-router';

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

var HibernatingRoute = function HibernatingRoute(props) {
  // This component is used ONLY as an indicator to HibernatingSwitch that it needs to keep this subtree around
  // even after the element itself unmounts. HibernatingSwitch will replace each <HibernatingRoute> with a standard
  // <Route> (with the extra wiring to make hibernation happen) -- so this component should never actually render.
  //
  // If this ever does render then it means the component wasn't swapped out -- so it's not being used properly.
  if (process.env.NODE_ENV !== 'production') {
    console.warn('HibernatingRoute is being used as a standalone component, instead of as a child of HibernatingSwitch. This will not do anything.', props);
  }

  return React.createElement(Route, Object.assign({}, props));
};

// https://github.com/ReactTraining/react-router/blob/f20cd62495c7bc4f8949ae6d28afe587f62b557a/packages/react-router/modules/Route.js#L56-L72

var renderRoute = function renderRoute(routerProps, routeProps) {
  var children = routeProps.children,
      component = routeProps.component,
      render = routeProps.render;
  return routerProps.match ? children ? typeof children === 'function' ? process.env.NODE_ENV !== 'production' ? children(routerProps) || null : children(routerProps) : children : component ? // eslint-disable-next-line @typescript-eslint/no-explicit-any
  React.createElement(component, routerProps) : render ? render(routerProps) : null : typeof children === 'function' ? process.env.NODE_ENV !== 'production' ? children(routerProps) || null : children(routerProps) : null;
};

var HibernatingSwitch = function HibernatingSwitch(_ref) {
  var children = _ref.children,
      maxCacheSize = _ref.maxCacheSize,
      maxCacheTime = _ref.maxCacheTime,
      WrapperComponent = _ref.WrapperComponent,
      allOtherProps = _objectWithoutPropertiesLoose(_ref, ["children", "maxCacheSize", "maxCacheTime", "WrapperComponent"]);

  var portalRecordCache = useMemo(function () {
    return LimitedCache({
      maxCacheSize: maxCacheSize,
      maxCacheTime: maxCacheTime
    });
  }, [maxCacheSize, maxCacheTime]);
  var currentPathKeyRef = useRef();
  var currentPathKey = currentPathKeyRef.current;

  var _useState = useState(null),
      currentPortalRecord = _useState[0],
      setCurrentPortalRecord = _useState[1];

  if (process.env.NODE_ENV !== 'production') {
    // eslint-disable-next-line react-hooks/rules-of-hooks
    var InitialWrapperComponentRef = useRef(WrapperComponent);

    if (WrapperComponent !== InitialWrapperComponentRef.current) {
      console.warn('The WrapperComponent component changed between renders: this will cause a remount');
      InitialWrapperComponentRef.current = WrapperComponent;
    }
  }

  var activatePortalForComponent = function activatePortalForComponent(routerProps, routeProps) {
    // We don't really care about the *path* that was matched. Instead, we care about the *url* that activated us
    var pathKey = routerProps.match.url;

    if (pathKey === currentPathKey) {
      // Nothing has really changed: just update any props
      currentPortalRecord.routerProps = routerProps;
      currentPortalRecord.routeProps = routeProps;
    } else {
      // New route! Stash the old and move to the new
      // But first, pull any previous portal record for the 'new' screen (in case it's in the cache) --
      // otherwise it might get removed when we add the old portal record, if the cache size is small.
      var previousPortalRecord = portalRecordCache.get(pathKey);

      if (currentPathKey) {
        portalRecordCache.set(currentPathKey, currentPortalRecord);
      }

      var newPortalRecord;

      if (previousPortalRecord) {
        // Reactivate the prior subtree
        previousPortalRecord.routerProps = routerProps;
        previousPortalRecord.routeProps = routeProps;
        newPortalRecord = previousPortalRecord;
      } else {
        // Make a new portal for the new subtree
        newPortalRecord = {
          portalNode: createHtmlPortalNode(),
          routerProps: routerProps,
          routeProps: routeProps
        };
      }

      currentPathKeyRef.current = pathKey;
      setCurrentPortalRecord(newPortalRecord);
    }
  };

  var deactivatePortal = function deactivatePortal() {
    if (currentPathKey && currentPortalRecord) {
      portalRecordCache.set(currentPathKey, currentPortalRecord); // We need to unmount the old OutPortal *immediately* so that it can swap in its original dom node,
      // or else React hits an error trying to unmount in the incorrect node

      currentPortalRecord.portalNode.unmount();
      currentPathKeyRef.current = '';
      setCurrentPortalRecord(null);
    }
  };

  var childrenWithHibernation = React.Children.map(children, function (child) {
    if (isElement(child)) {
      var type = child.type,
          routeProps = child.props,
          _child$props = child.props,
          isHibernatingRoute = _child$props.isHibernatingRoute,
          allOtherRouteProps = _objectWithoutPropertiesLoose(_child$props, ["children", "component", "isHibernatingRoute", "render"]);

      if (type === HibernatingRoute || isHibernatingRoute) {
        // Replace it: it will activate the remote node when the route matches
        return React.createElement(Route, Object.assign({}, allOtherRouteProps, {
          render: function render(routerProps) {
            activatePortalForComponent(routerProps, routeProps);
            return null;
          }
        }));
      }

      if (type !== Redirect && isHibernatingRoute !== false) {
        // Every child should either be a Hibernating Route, vanilla Route, or Redirect -- and if it's neither a
        // vanilla route nor a Hibernating Route, it's probably a custom wrapper around Route.
        // If it's any non-Hibernating route Route then we need to deactivate any portal which might be active from
        // a Hibernating Route. But no matter what it might be, we want to deactivate before rendering, to be safe.
        return React.createElement(Route, Object.assign({}, allOtherRouteProps, {
          render: function render(routerProps) {
            deactivatePortal();
            return renderRoute(routerProps, routeProps);
          }
        }));
      }
    } // If it's a Redirect, or anything we can't recognize, let it pass through as-is


    return child;
  });
  var portalRecordCacheFull = portalRecordCache.getAll();
  var allPortalKeys = Object.keys(portalRecordCacheFull);

  if (currentPathKey && !allPortalKeys.includes(currentPathKey)) {
    allPortalKeys.push(currentPathKey);
  }

  return React.createElement(React.Fragment, null, React.createElement(Switch, Object.assign({}, allOtherProps), childrenWithHibernation), React.createElement(React.Fragment, null, allPortalKeys.map(function (pathKey) {
    var portalRecord = pathKey === currentPathKey ? currentPortalRecord : portalRecordCacheFull[pathKey];

    if (!portalRecord) {
      if (process.env.NODE_ENV !== 'production') {
        console.warn("portalRecord is missing for pathKey \"" + pathKey + "\"");
      }

      return null;
    }

    var portalNode = portalRecord.portalNode,
        routerProps = portalRecord.routerProps,
        routeProps = portalRecord.routeProps;
    var routeContent = renderRoute(routerProps, routeProps);
    var wrappedRouteContent = WrapperComponent ? React.createElement(WrapperComponent, {
      shouldUpdate: pathKey === currentPathKey
    }, routeContent) : routeContent;
    return React.createElement(InPortal, {
      key: pathKey,
      node: portalNode
    }, wrappedRouteContent);
  }), !!currentPortalRecord && React.createElement(OutPortal, {
    node: currentPortalRecord.portalNode
  })));
};

HibernatingSwitch.defaultProps = {
  maxCacheSize: 5,
  maxCacheTime: 5 * 60 * 1000,
  WrapperComponent: null
};

export { HibernatingRoute, HibernatingSwitch };
//# sourceMappingURL=react-router-hibernate.esm.js.map
